#include "cquad.h"

double cquad_dqaws(double f(), double a, double b, double alfa, double beta, int integr, double epsabs, double epsrel, double *abserr, int *neval, int *ier, int *last, int *iwork, double *work)
{
/******************************************************************************
***  Date written:         03/31/2007  (mm/dd/yyyy)                         ***
***-------------------------------------------------------------------------***
***  Last revision date:   11/02/2012  (mm/dd/yyyy)                         ***
***-------------------------------------------------------------------------***
***  Author:    Gianluca Russo    (gianlurusso@fastwebnet.it)               ***
***             Salvatore Cuomo   (salvatore.cuomo@unina.it)                ***
***                                                                         ***
***             University of Naples "Federico II"                          ***
***             Dipartimento di Matematica e Applicazioni "R. Caccioppoli"  ***
***									    ***
*** 									    ***
***									    ***
*** This code is provided "as is", without any warranty of any kind,        ***
*** either expressed or implied, including but not limited to, any implied  ***
*** warranty of merchantibility or fitness for any purpose. In no event     ***
*** will any party who distributed the code be liable for damages or for    ***
*** any claim(s) by any other party, including but not limited to, any      ***
*** lost profits, lost monies, lost data or data rendered inaccurate,	    ***
*** losses sustained by third parties, or any other special, incidental or  ***	
*** consequential damages arising out of the use or inability to use the    ***
*** program, even if the possibility of such damages has been advised	    ***	
*** against. The entire risk as to the quality, the performance, and the    ***	
*** fitness of the program for any particular purpose lies with the party   ***
*** using the code.							    ***
***									    ***
*** This code, and any derivative of this code, may not be used in a        ***
*** commercial package without the prior explicit written permission of     ***
*** the authors. Verbatim copies of this code may be made and distributed   ***
*** in any medium, provided that this copyright notice is not removed or    ***
*** altered in any way. No fees may be charged for distribution of the      ***
*** codes, other than a fee to cover the cost of the media and a            ***
*** reasonable handling fee.						    ***
***-------------------------------------------------------------------------***
***  Purpose:                                                               ***
***                                                                         ***
***    The function provides a C interface to the fortran routine dqaws     ***
***    of quadpack package.                                                 ***
***                                                                         ***
***    The function calculates an approximation result to a given definite  ***
***    integral i = integral of f(x) * w (x) over (a, b), where w shows a   ***
***    singular behaviour at the end points (see parameter integr),         *** 
***    hopefully satisfying following claim for accuracy:                   ***
***                                                                         ***
***              abs(i - result) <= max(epsabs, epsrel * abs(i))            ***
***-------------------------------------------------------------------------***
***  Description:                                                           ***
***                                                                         ***
***    Integrates f(x) * w(x) over (a, b) with a < b finite, and            ***
***    w(x) = ((x - a)^alfa) * ((b-x)^beta) * v(x) where:                   ***
***                                                                         ***
***                        v(x) = 1 or                                      ***
***                        v(x) = LOG(x - a) or                             ***
***                        v(x) = LOG(b - x) or                             ***
***                        v(x) = LOG(x - a) * LOG(b - x)                   ***
***                                                                         ***
***    and alfa > (-1), beta > (-1).                                        ***
***                                                                         ***
***    The user specifies a, b, alfa, beta and the type of the function v.  ***
***    A globally adaptive subdivision strategy is applied, with modified   ***
***    Clenshaw-Curtis integration on the subintervals which contain a      ***
***    or b.                                                                ***
***                                                                         ***
***    Double precision version.                                            ***
***-------------------------------------------------------------------------***
***  Input:                                                                 ***
***                                                                         ***
***    f        function subprogram defining the integrand function f(x)    ***
***                                                                         ***
***    a        lower limit of integration                                  ***
***                                                                         ***
***    b        upper limit of integration (b > a)                          ***
***                                                                         ***
***    alfa     parameter in the integrand function (alfa > -1)             ***
***                                                                         ***
***    beta     parameter in the integrand function (beta > -1)             ***
***                                                                         ***
***    integr   indicates which weight functions is to be used:             ***
***                                                                         ***
***             integr = 1 --> (x - a)^alfa * (b - x)^beta                  ***
***             integr = 2 --> (x - a)^alfa * (b - x)^beta * log(x - a)     ***
***             integr = 3 --> (x - a)^alfa * (b - x)^beta * log(b - x)     ***
***             integr = 4 --> (x - a)^alfa * (b - x)^beta * log(x - a) *   ***
***                            log(b - x)                                   ***
***                                                                         ***
***    epsabs   absolute accoracy requested                                 ***
***                                                                         ***
***    epsrel   relative accoracy requested                                 ***
***-------------------------------------------------------------------------***
***  Output:                                                                ***
***                                                                         ***
***    res      approximation to the integral                               ***
***                                                                         ***
***    abserr   estimate of the modulus of the absolute error               ***
***                                                                         ***
***    neval    number of integrand evaluations                             ***
***                                                                         ***
***    ier      normal termination flag:                                    ***
***                                                                         ***
***               ier = 0 --> normal and reliable termination of the        ***
***                           function. It is assumed that the requested    ***
***                           accuracy has been achieved.                   ***
***                                                                         ***
***               ier = 1 --> maximum number of subdivisions allowed has    ***
***                           been achieved.                                ***
***                                                                         ***
***               ier = 2 --> the occurrence of roundoff error is detected  ***
***                           which prevents the requested tolerance from   ***
***                           being achieved.                               ***
***                                                                         ***
***               ier = 3 --> extremely bad integrand behaviour occurs at   ***
***                           some points of the integration interval.      ***
***                                                                         ***
***               ier = 6 --> the input is invalid because b <= a, or       ***
***                           alfa <= -1, or beta <= -1, or integr < 1 or   ***
***                           integr > 4, or (epsabs <= 0 and epsrel < max( ***
***                           50*rel.mach.acc.,0.5d-28)), or limit < 2, or  ***
***                           lenw < limit * 4.                             ***
***                           In all these events res, abserr, neval and    ***
***                           last are set to zero.                         ***
***                                                                         ***
***    last     number of subintervals produced in the subdivision process  ***
***-------------------------------------------------------------------------***
***  Workspace:                                                             ***
***                                                                         ***
***    limit    dimensionin value for iwork.                                ***
***             limit determines the maximum number of subintervals in      ***
***             the partition of the given integration interval (a,b)       ***
***             (limit > 1)                                                 ***
***                                                                         ***
***    lenw     dimensioning value for work.                                ***
***             (lenw must be at least limit * 4)                           ***
***                                                                         ***
***    iwork    vector of dimension at least limit, the first k elements    ***
***             of which contain pointers to the error estimates over the   ***
***             subintervals, such that work(limit * 3 + iwork(1)), ...,    ***
***             work(limit * 3 + iwork(k)) form a decreasing sequence, with ***
***             k = last if last <= (limit/2 + 2), and k = limit + 1 - last ***
***             otherwise                                                   *** 
***                                                                         ***
***    work     vector of dimension at least lenw.                          ***
***             On return:                                                  ***
***                work(1), ..., work(last) contain the left end points of  ***
***                  the subintervals in the partition of (a,b)             ***
***                work(limit + 1), ..., work(limit + last) contain the     ***
***                  right end points                                       ***
***                work(limit * 2 + 1), ..., work(limit * 2 + last) contain ***
***                  the integral approximations over the subintervals      ***
***                work(limit * 3 + 1), ..., work(limit * 3 + last) contain ***
***                  the error estimates                                    ***
***-------------------------------------------------------------------------***
***  References:      QUADPACK - Piessens, de Doncker, Uberhuber, Kahaner   ***
***                                                                         ***
***                   http://netlib.org                                     ***
******************************************************************************/

    int lenw = LENW, limit = LIMIT;
    double res;

    dqaws_(f, &a, &b, &alfa, &beta, &integr, &epsabs, &epsrel, &res, abserr, neval, ier, &limit, &lenw, last, iwork, work);

    return res;
}
